/*:
* @plugindesc This plugin logs text event meta data, which can be used to cut recordings and to create voice acting mods for TCoAaL.
* @author canopy 
*
* @help
* The logging will take place automatically while this plugin is active.
* To process audio files copy them and the /tomb/mods/voiceactingcuttingtool_data/sessions/session<number>.json 
* you want to process into the /tomb/mods/voiceactingcuttingtool_data/process_sessions folder and
* configure the session<number>.json to your needs.
* Afterwards press the Cut Rec button in the main menu of the game.
* More info can be found in the repo of this mod: https://codeberg.org/canopy/voiceactingcuttingtool
*/

(function() {    
    const fs = require('fs');
    const path = require('path');
    const { exec } = require('child_process');

    const rootPath = process.cwd().replace(/\\/g, "/");
    const dataSubPath = "/tomb/mods/voiceactingcuttingtool_data";
    const sessionsSubPath = dataSubPath + "/sessions";
    const processSessionsSubPath = dataSubPath + "/process_audio";
    const processSessionsOutputSubPath = processSessionsSubPath + "/output";
    const processSessionsModsSubPath = "/tomb/mods/";
    const processAudioConfigPath = path.posix.join(rootPath, processSessionsSubPath, "processAudio.config");
    const modCreationConfigPath = path.posix.join(rootPath, processSessionsSubPath, "modCreation.config");
    const modCreationLogPath = path.posix.join(rootPath, processSessionsSubPath, "modCreation.log");
    createSubDir(sessionsSubPath);
    createSubDir(processSessionsOutputSubPath);
    writeDataModJson();
    createProcessConfig();
    createModDataConfig();

    const sessionLogFilePath = rootPath + "/" + sessionsSubPath + "/" + getNextSessionFile(rootPath + "/" + sessionsSubPath);

    let requestLogFileWrite = false;
    let isWritingLogFile = false;

    let sessionLog = initSessionLog();
    let currentEventStartTime;
    let currentTextEventId = -1;
    ///* FUNCTIONS *///

    function initSessionLog() {
        return {
            gameStartTime: new Date(),
            textEvents: []
        };
    }

    async function createProcessConfig(){
        if (fs.existsSync(processAudioConfigPath)) {return;}
        
        const processAudioJson = {
            sessions:[
                {
                    name: "session1",
                    recordings: [
                        {
                            fileName: "replaceThisWithYourRecording.mkv",
                            timeOffset: "00:00:00.000"
                        },
                        {
                            fileName: "aDifferentRecordingFromTheSameSession.mp4",
                            timeOffset: "00:00:00.000",
                            options: "-b:a 128k",
                            labelsToProcess: [
                                "Narrator", 
                                "Ashley"
                            ]
                        }
                    ]
                }
            ],
            log: ["FILEWRITTEN"], //["FILEWRITTEN", "DEBUG"]
        }
        try {
            const jsonString = JSON.stringify(processAudioJson, null, 2);
            await fs.writeFile(processAudioConfigPath, jsonString, 'utf-8');
        } catch (err) {
            console.error('Error writing JSON file:', err);
        }
    }

    async function createModDataConfig(){
        if (fs.existsSync(modCreationConfigPath)) {return;}
        
        const modCreationJson = {
            dataForModJson:{
                id: "my_va_mod",
                name: "My 1st VA mod",
                authors: [
                    "Your Name"
                ],
                description: "Enter a description here!",
                version: "1.0.0",
                dependencies: {
                    "game": "*",
                    "spec": "0.1.0",
                    "voiceacting": ">=0.4.0"
                },
            },
            modSubFolderForVoiceActingFiles: "voiceacting/english",
            audioFileSources:[
                {
                    folder: "output/session1/replaceThisWithYourRecording"
                },
                {
                    folder: "output/session1/aDifferentRecordingFromTheSameSession",
                    subFolders: [
                        "Narrator", 
                        "Ashley"
                    ]
                },
                {
                    folder: "manualReplacements",
                    subFolders: [
                        "Ashley"
                    ]
                }
            ]
        }
        try {
            const jsonString = JSON.stringify(modCreationJson, null, 2);
            await fs.writeFile(modCreationConfigPath, jsonString, 'utf-8');
        } catch (err) {
            console.error('Error writing JSON file:', err);
        }
    }

    async function writeDataModJson(){
        const modDataJsonPath = path.posix.join(rootPath, dataSubPath, "mod.json");
        if (fs.existsSync(modCreationConfigPath)) {return;}

        const dataModJson = {
            "id": "voiceactingcuttingtool_data",
            "name": "voiceactingcuttingtool_data",
            "authors": [
                "Canopy"
            ],
            "description": "This mod is created by the voiceactingcuttingtool mod and is used to store session logs and data to process.",
            "version": "1.0.0",
            "dependencies": {
                "game": "*",
                "spec": "0.1.0"
            },
            "files": {}
        }
        try {
            const jsonString = JSON.stringify(dataModJson, null, 2);
            await fs.writeFile(modDataJsonPath, jsonString, 'utf-8');
        } catch (err) {
            console.error('Error writing JSON file:', err);
        }
    }


    function createSubDir(newDir) {    
        let subDirs = newDir.split("/");
        if (subDirs.length < 1) {throw new Error("Invalid folder name: Folder names cannot be empty");}
        let folderPath = rootPath;
        for(let folderName of subDirs){
            folderPath = (folderPath + "/" + folderName);
            if (!fs.existsSync(folderPath)) {
            fs.mkdirSync(folderPath);
            }
        }
    } 

    function getNextSessionFile(folderPath, increment = 1) {
        const files = fs.readdirSync(folderPath);
        const sessionPattern = /^session(\d+)\.json$/;
        let sessionNumbers = [];
        files.forEach(file => {
            const match = file.match(sessionPattern);
            if (match) {
                sessionNumbers.push(parseInt(match[1], 10));
            }
        });
        if (sessionNumbers.length === 0) { return 'session1.json';}
        const highestSessionNumber = Math.max(...sessionNumbers);
        return `session${highestSessionNumber + increment}.json`;
    }

    async function writeSessionLogIntoFile() {
        try {
          const jsonString = JSON.stringify(sessionLog, null, 2);
          await fs.writeFile(sessionLogFilePath, jsonString, 'utf-8');
          if (requestLogFileWrite){
            requestLogFileWrite = false;
            writeSessionLogIntoFile();
          } else {
            isWritingLogFile = false;
          }
        } catch (err) {
          console.error('Error writing JSON file:', err);
        }
    }

    function formatTimeDifference(timeDifferenceInMs) {
        const isNegative = timeDifferenceInMs < 0;
        timeDifferenceInMs = Math.abs(timeDifferenceInMs);

        let hours = Math.floor(timeDifferenceInMs / (1000 * 60 * 60));
        let minutes = Math.floor((timeDifferenceInMs % (1000 * 60 * 60)) / (1000 * 60));
        let seconds = Math.floor((timeDifferenceInMs % (1000 * 60)) / 1000);
        let milliseconds = timeDifferenceInMs % 1000;
    
        hours = String(hours).padStart(2, '0');
        minutes = String(minutes).padStart(2, '0');
        seconds = String(seconds).padStart(2, '0');
        milliseconds = String(milliseconds).padStart(3, '0');

        let timeString = `${hours}:${minutes}:${seconds}.${milliseconds}`;
        return isNegative ? `-${timeString}` : timeString;
    }

    function parseTimeDifference(timeDifferenceString) {
        const isNegative = timeDifferenceString.startsWith('-');
        if (isNegative) {timeDifferenceString = timeDifferenceString.slice(1);}
    
        //HH:MM:SS.sss; MM:SS.sss and SS.sss formats are supported
        const timePattern = /^(?:(\d{2}):)?(?:(\d{2}):)?(\d{2})\.(\d{3})$/;
        const match = timeDifferenceString.match(timePattern);
        if (!match) {throw new Error('Invalid time string format');}
    
        let hours = 0, minutes = 0, seconds = 0, milliseconds = 0;
        if (match[2]) {
            hours = parseInt(match[1], 10);
            minutes = parseInt(match[2], 10);
            seconds = parseInt(match[3], 10);
            milliseconds = parseInt(match[4], 10);
        } else if (match[1]) {
            minutes = parseInt(match[1], 10);
            seconds = parseInt(match[3], 10);
            milliseconds = parseInt(match[4], 10);
        } else if (match[3]) {
            seconds = parseInt(match[3], 10);
            milliseconds = parseInt(match[4], 10);
        } else {
            throw new Error('unexpected matches in time string format');
        }
    
        let totalMilliseconds =
            (hours * 60 * 60 * 1000) +
            (minutes * 60 * 1000) +
            (seconds * 1000) +
            milliseconds;
    
        return isNegative ? -totalMilliseconds : totalMilliseconds;
    }

    function cutAudio(session, createLabelFolders, logStream, logTypes){
        createLabelFolders = createLabelFolders || true;
        session.recordings.forEach(recording => {
            // Ensure the output folder exists
            const outputBaseFolderSubPath = path.posix.join(processSessionsOutputSubPath, session.name, path.basename(recording.fileName, path.extname(recording.fileName)));
            const outputBaseFolderPath = path.posix.join(rootPath, outputBaseFolderSubPath);
            if (!fs.existsSync(outputBaseFolderPath)) {
                createSubDir(outputBaseFolderSubPath);
            } else {
                logMessage(logStream, true, "Output folder \"" + outputBaseFolderSubPath + "\" already exists and will NOT be overridden. \n Delete it manually before you try again!");
                return;
            }
            sessionLog.textEvents.forEach(segment => {
                if (segment.linesKey == null){return;}
                let labelText = segment.labelText || "Narrator";
                labelText = labelText
                    .replace(/\?/g, '_QUESTIONMARK_')
                    .replace(/ /g, '_');
                if (recording.labelsToProcess !== undefined){
                    let textEventNeedsToBeProcessed = false;
                    for (let labelToProcess of recording.labelsToProcess){
                        if (labelToProcess == labelText){
                            textEventNeedsToBeProcessed = true;
                        }
                    }
                    if (!textEventNeedsToBeProcessed){return;}
                } //else process all labels

                const outputFileExtension = recording.outputFileExtension || ".ogg";
                let outputFileName;
                if (outputFileExtension.startsWith('.')){
                    outputFileName = segment.linesKey + outputFileExtension;
                } else {
                    outputFileName = segment.linesKey + "." + outputFileExtension;
                }
                let outputFilePath;

                if (createLabelFolders){
                    const outputFolderPath = path.posix.join(outputBaseFolderPath, labelText);
                    if (!fs.existsSync(outputFolderPath)) {
                        createSubDir(path.posix.join(outputBaseFolderSubPath, labelText));
                    }
                    outputFilePath = path.posix.join(outputBaseFolderPath, labelText, outputFileName);
                } else {
                    outputFilePath = path.posix.join(outputBaseFolderPath, outputFileName);
                }

                const timeOffset = recording.timeOffset || "00:00:00.000";

                const inputFilePath = path.posix.join(rootPath, processSessionsSubPath, recording.fileName);
                const timeStart = formatTimeDifference(parseTimeDifference(segment.timeStart) + parseTimeDifference(timeOffset));
                const timeDuration = segment.timeDuration;
                const options = recording.options || "";
                const command = `ffmpeg -i "${inputFilePath}" -vn -ss ${timeStart} -t ${timeDuration} -acodec libvorbis ${options} "${outputFilePath}"`;
                exec(command, (error, stdout, stderr) => {
                    if (error) {
                        console.error(`Error cutting audio: ${error.message}`);
                        return;
                    }
                    if (stderr) {
                        logMessage(logStream, "DEBUG", `ffmpeg stderr: ${stderr}`, logTypes);
                    }
                    logMessage(logStream, "FILEWRITTEN", `${outputFilePath} created.`, logTypes);
                });
            });
        });
    }

    function logMessage(logStream, logType, message, logTypesToPrint){
        if (logType === true || (logTypesToPrint != undefined && logTypesToPrint.includes(logType))){
            logStream.write(message + "\n");
        }
    }

    ///* HOOKS *///

    // on start of text event
    var CAN_VoiceActingCuttingTool_Game_Interpreter_command101 = Game_Interpreter.prototype.command101;
    Game_Interpreter.prototype.command101 = function(){
        CAN_VoiceActingCuttingTool_Game_Interpreter_command101.call(this);
        let mvText = this._list[this._index-1].parameters[0];
        let linesKeyMatch = mvText.match(/\(lines\)\[([^\]]+)\]/);
        let labelTextMatch = $gameMessage._texts[0].match(/\\n<([^>]+)>/);
        let linesText = null;
        if ($gameMessage._texts.length > 1){
            linesText = $gameMessage._texts.slice(1);
        }
        currentEventStartTime = new Date();
        sessionLog.textEvents.push(
            {
                linesKey: linesKeyMatch ? linesKeyMatch[1] : null,
                labelText: labelTextMatch ? labelTextMatch[1] : null,
                linesText: linesText,
                timeStart: formatTimeDifference(currentEventStartTime - sessionLog.gameStartTime),
                timeDuration: null
            }
        )
        currentTextEventId += 1;
    }

   // on end of text event
    var CAN_VoiceActingCuttingTool_Window_Message_terminateMessage = Window_Message.prototype.terminateMessage;
    Window_Message.prototype.terminateMessage = function(){
        CAN_VoiceActingCuttingTool_Window_Message_terminateMessage.call(this);
        const durationOfEvent = formatTimeDifference(new Date() - currentEventStartTime);
        sessionLog.textEvents[currentTextEventId].timeDuration = durationOfEvent;
    }

    // on end of event page
    var CAN_VoiceActingCuttingTool_Game_Interpreter_executeCommand = Game_Interpreter.prototype.executeCommand;
    Game_Interpreter.prototype.executeCommand = function(){
        if (this._index == this._list.length){
            if (this._list.some(item => item["code"] == 401)){
                console.log("sessionLog: ", sessionLog);
                if (isWritingLogFile){
                    requestLogFileWrite;
                } else {
                    isWritingLogFile = true;
                    writeSessionLogIntoFile();
                }
            }
        }
        CAN_VoiceActingCuttingTool_Game_Interpreter_executeCommand.call(this);
    }
    

    //* Menu Botton to Cut VA recordings *//
    var CAN_VoiceActingCuttingTool_WindowTitleCommand_makeCommandList = Window_TitleCommand.prototype.makeCommandList;
    Window_TitleCommand.prototype.makeCommandList = function() {
        CAN_VoiceActingCuttingTool_WindowTitleCommand_makeCommandList.call(this);
        this.addCommand("Cut Rec", 'cutRecording');
        this.addCommand("CutsToMod", 'createVoiceActingMod');    
    };

    var CAN_VoiceActingCuttingTool_Scene_Title_createCommandWindow = Scene_Title.prototype.createCommandWindow;
    Scene_Title.prototype.createCommandWindow = function() {
        CAN_VoiceActingCuttingTool_Scene_Title_createCommandWindow.call(this);
        this._commandWindow.setHandler('cutRecording', this.cutRecording.bind(this));
        this._commandWindow.setHandler('createVoiceActingMod', this.createVoiceActingMod.bind(this));
    };

    Scene_Title.prototype.cutRecording = function() {
        const processAudioConfig = JSON.parse(fs.readFileSync(processAudioConfigPath, 'utf-8'));
        processAudioConfig.sessions.forEach(session =>{
            const sessionFileNameToProcess = session.name + ".json";
            const logStream = fs.createWriteStream(rootPath + "/" + processSessionsSubPath + "/" + session.name + ".log", { flags: 'a' });

            let sessionLogPath = path.posix.join(rootPath, processSessionsSubPath, sessionFileNameToProcess);
            if (!fs.existsSync(sessionLogPath)) {
                sessionLogPath = path.posix.join(rootPath, sessionsSubPath, sessionFileNameToProcess);
            }
            sessionLog = JSON.parse(fs.readFileSync(sessionLogPath, 'utf-8'));

            cutAudio(session, processAudioConfig.createLabelFolders, logStream, processAudioConfig.log);
        });
        
        sessionLog = initSessionLog();

        this._commandWindow.close();
        SceneManager.push(Scene_Title);
    }

    Scene_Title.prototype.createVoiceActingMod = function() {
        const logStream = fs.createWriteStream(modCreationLogPath, { flags: 'a' });        
        try {
            let modCreationConfig = JSON.parse(fs.readFileSync(modCreationConfigPath, 'utf-8'));
            const modFolderPath = path.posix.join(rootPath, processSessionsModsSubPath, modCreationConfig.dataForModJson.id);
            if (fs.existsSync(modFolderPath)) {
                const message = "Error: modfolder \"" + modCreationConfig.dataForModJson.id + "\" already exists.\n " +
                    "Delete \"" + modFolderPath + "\" manually if you want to build the mod again!";
                logMessage(logStream, true, message);
                this._commandWindow.close();
                SceneManager.push(Scene_Title);
                return;
            }
            const modVoiceActingSubFolder = path.posix.join(processSessionsModsSubPath, modCreationConfig.dataForModJson.id, modCreationConfig.modSubFolderForVoiceActingFiles);
            //console.log(modVoiceActingFolder);
            createSubDir(modVoiceActingSubFolder);
            const modVoiceActingFolderPath = path.posix.join(rootPath, modVoiceActingSubFolder);
            modCreationConfig.audioFileSources.forEach(audioSource =>{
                let folders = [];
                if (audioSource.subFolders == undefined){
                    folders.push(path.posix.join(rootPath, processSessionsSubPath, audioSource.folder));
                } else {
                    audioSource.subFolders.forEach(subFolder =>{
                        folders.push(path.posix.join(rootPath, processSessionsSubPath, audioSource.folder, subFolder));
                    });
                }
                folders.forEach(srcDir =>{
                    copyOggFilesSync(srcDir, modVoiceActingFolderPath, logStream);
                })
            });

            const filesInVAFolder = fs.readdirSync(modVoiceActingFolderPath);
            const oggFiles = filesInVAFolder.filter(file => {
                const filePath = path.posix.join(modVoiceActingFolderPath, file);
                const stat = fs.lstatSync(filePath);
                return stat.isFile() && path.extname(file) === '.ogg';  // Check if it's a .ogg file
            });
            const assetList = oggFiles.map(name => modCreationConfig.modSubFolderForVoiceActingFiles + "/" + name);
            modCreationConfig.dataForModJson.files = {assets: assetList};

            let modJson = JSON.stringify(modCreationConfig.dataForModJson, null, 2);
            fs.writeFileSync(modFolderPath + "/mod.json", modJson, 'utf-8');
        } catch (err) {
            console.log(`Error: ${err.message}`);
            logMessage(logStream, true, `Error: ${err.message}`);
        }

        function copyOggFilesSync(srcDir, destDir, logStream) { 
            const files = fs.readdirSync(srcDir);
            for (const file of files) {
                const srcPath = path.posix.join(srcDir, file);
                const destPath = path.posix.join(destDir, file);
                const stat = fs.lstatSync(srcPath);
                if (stat.isDirectory()) {
                    copyOggFilesSync(srcPath, destDir, logStream);
                } else if (stat.isFile() && path.extname(file) === '.ogg') {
                    fs.copyFileSync(srcPath, destPath);
                    logMessage(logStream, true, `Copied: ${srcPath} -> ${destPath}`);
                }
            }
        }

        this._commandWindow.close();
        SceneManager.push(Scene_Title);
    }
})();
